package scd.correlation;

import java.io.File;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.PriorityQueue;
import java.util.Scanner;

import scd.constants.Constants;

public class Correlation {
	public static ArrayList<Pair> extractPairs(ArrayList<double[]> data) throws Exception {
		int k = data.get(0).length;
		double[][] corr = pairwiseScores(data);

		// identify the dimension with largest sum of correlation scores
		double[] C = new double[k];
		int[] E = new int[k];
		int first = 0;
		double maxScore = Double.NEGATIVE_INFINITY;
		for (int i = 0; i < k; i++) {
			C[i] = Double.NEGATIVE_INFINITY;
			E[i] = -1;
			
			double sum = 0;
			for (int j = 0; j < k; j++) {
				sum += corr[i][j];
			}

			if (sum > maxScore) {
				maxScore = sum;
				first = i;
			}
		}
		
		E[first] = -1;
		C[first] = Double.POSITIVE_INFINITY;
		PriorityQueue<Element> pq = new PriorityQueue<Element>(Collections.reverseOrder());
		Element tmp = null;
		for (int i = 0; i < k; i++) {
			if (i != first) {
				E[i] = first;
				C[i] = corr[i][first];
				tmp = new Element(i, C[i]);
				pq.offer(tmp);
			}
		}

		ArrayList<Pair> ret = new ArrayList<Pair>();
		Pair p = null;
		while (!pq.isEmpty()) {
			tmp = pq.poll();
			if (E[tmp.vertex] < tmp.vertex) {
				p = new Pair(E[tmp.vertex], tmp.vertex, corr[E[tmp.vertex]][tmp.vertex]);
			} else {
				p = new Pair(tmp.vertex, E[tmp.vertex], corr[E[tmp.vertex]][tmp.vertex]);
			}
			ret.add(p);
			
			ArrayList<Element> changedElements = new ArrayList<Element>();
			for (Element e : pq) {
				if (corr[tmp.vertex][e.vertex] > e.score) {
					E[e.vertex] = tmp.vertex;
					e.score = corr[tmp.vertex][e.vertex];
					changedElements.add(e);
				}
			}
			for (Element e : changedElements) {
				pq.remove(e);
				pq.offer(e);
			}
		}

		return ret;
	}

	private static double[][] pairwiseScores(ArrayList<double[]> data) throws Exception {
		int N = data.size();
		int k = data.get(0).length;
		ArrayList<double[][]> vectors = getVectors(N, Constants.S1,
				Constants.S2);
		double[][] u = vectors.get(0);
		double[][] w = vectors.get(1);
		double[] sum = computeVectorSum(w);

		double[][] corr = new double[k][k];
		for (int i = 0; i < k; i++) {
			double[] x = getColumn(data, i);
			Arrays.sort(x);
			for (int j = i + 1; j < k; j++) {
				double[] y = getColumn(data, j);
				Arrays.sort(y);
				corr[i][j] = correlationScore(x, y, u, w, sum);
				corr[j][i] = corr[i][j];
			}
		}

		return corr;
	}

	private static double[] getColumn(ArrayList<double[]> data, int col) {
		int N = data.size();
		double[] x = new double[N];
		for (int i = 0; i < N; i++) {
			x[i] = data.get(i)[col];
		}

		return x;
	}

	private static ArrayList<double[][]> getVectors(int N, int S1, int S2) throws Exception {
		Runtime rt = Runtime.getRuntime();
		String cmdFileName = "H:\\Development\\LARGESPACE\\supporting\\Sketch\\Debug\\Sketch.exe";
		String fileU = "u.txt";
		String fileW = "w.txt";
		String[] cmd = { cmdFileName, Integer.toString(S1), Integer.toString(S2), Integer.toString(N), fileU, fileW };
		Process p = rt.exec(cmd);
		int exit = p.waitFor();
		
		if (exit != 0) {
			throw new Exception("cannot generate sketch!");
		}
		
		double[][] u = readSketchVectors(N, S1, S2, fileU);
		double[][] w = readSketchVectors(N, S1, S2, fileW);
		ArrayList<double[][]> ret = new ArrayList<double[][]>();
		ret.add(u);
		ret.add(w);
		return ret;
	}
	
	private static double[][] readSketchVectors(int N, int S1, int S2, String file) throws Exception {
		int prod = S1 * S2;
		double[][] a = new double[N][prod];
		Scanner scan = new Scanner(new File(file));
		String[] vals = null;
		String plus = "+1";
		String minus = "-1";
		for (int i = 0; i < prod; i++) {
			vals = scan.nextLine().split(" ");
			for (int j = 0; j < N; j++) {
				if (vals[j].equals(plus)) {
					a[j][i] = 1;
				} else if (vals[j].equals(minus)) {
					a[j][i] = -1;
				} else {
					scan.close();
					throw new IllegalArgumentException(String.format("invalid sketch vector: %s", vals[j]));
				}
			}
		}
		scan.close();
		
		return a;
	}

	private static double[] computeVectorSum(double[][] w) {
		int rows = w.length;
		int cols = w[0].length;
		double[] sum = new double[cols];

		for (int j = 0; j < cols; j++) {
			sum[j] = 0;
			for (int i = 0; i < rows; i++) {
				sum[j] += w[i][j];
			}
		}

		return sum;
	}

	private static double correlationScore(double[] x, double[] y,
			double[][] u, double[][] w, double[] sum) {
		int N = x.length;
		int S2 = Constants.S2;
		int S1 = Constants.S1;
		double MAXVAL = Constants.MAXVAL;

		double[] scores = new double[S2];
		for (int s2 = 0; s2 < S2; s2++) {
			scores[s2] = 0;
			for (int s1 = 0; s1 < S1; s1++) {
				double s = sum[s1 + s2 * S1];
				double T1 = 0;
				double T1x = 0;
				double T1xx = 0;
				double T1y = 0;
				double T1yy = 0;
				double T2 = 0;
				double T2x = 0;
				double T2xx = 0;
				double T2y = 0;
				double T2yy = 0;
				double T3 = 0;
				double T3x = 0;
				double T3xx = 0;
				double T3y = 0;
				double T3yy = 0;
				for (int i = 0; i < N; i++) {
					s -= w[i][s1 + s2 * S1];

					T1xx += (MAXVAL - x[i]) * w[i][s1 + s2 * S1];
					T1x += (T1xx + (MAXVAL - x[i]) * s) * u[i][s1 + s2 * S1];
					T1yy += (MAXVAL - y[i]) * w[i][s1 + s2 * S1];
					T1y += (T1yy + (MAXVAL - y[i]) * s) * u[i][s1 + s2 * S1];

					T2xx += (MAXVAL - x[i]);
					T2x += (T2xx + (MAXVAL - x[i]) * (N - i - 1))
							* u[i][s1 + s2 * S1];
					T2yy += (MAXVAL - y[i]);
					T2y += (T2yy + (MAXVAL - y[i]) * (N - i - 1))
							* u[i][s1 + s2 * S1];

					T3xx += (MAXVAL - x[i]);
					T3x += T3xx + (MAXVAL - x[i]) * (N - i - 1);
					T3yy += (MAXVAL - y[i]);
					T3y += T3yy + (MAXVAL - y[i]) * (N - i - 1);
				}

				T1 = T1x * T1y;
				T2 = T2x * T2y;
				T3 = T3x * T3y;
				scores[s2] += T1 / (N * N) - 2 * T2 / Math.pow(N, 3) + T3
						/ Math.pow(N, 4);
				scores[s2] = scores[s2] / s1;
			}
		}

		Arrays.sort(scores);
		if (S2 % 2 == 1) {
			return scores[S2 / 2];
		} else {
			return (scores[S2 / 2 - 1] + scores[S2 / 2]) / 2;
		}
	}
}