package scd.transform;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.PriorityQueue;

import org.apache.commons.math3.linear.*;

import scd.common.CommonUtils;
import scd.constants.Constants;

public class Transform {
	private static double[][] getData(ArrayList<double[]> data) {
		double[][] a = new double[data.size()][];
		for (int i = 0; i < a.length; i++) {
			a[i] = data.get(i);
		}

		return a;
	}

	public static RealMatrix[] pca(ArrayList<double[]> data) {
		RealMatrix A = MatrixUtils.createRealMatrix(getData(data));
		int n = A.getColumnDimension();
		EigenDecomposition eigenDecomp = new EigenDecomposition(A.transpose().multiply(A));
		RealMatrix V = eigenDecomp.getV();
		
		// identify k
		int k = identifyK(eigenDecomp.getD(), n);

		// create results
		RealMatrix[] ret = new RealMatrix[2];
		ret[1] = V.getSubMatrix(0, n - 1, 0, k - 1);
		ret[0] = A.multiply(ret[1]);
		return ret;
	}

	public static RealMatrix[] fastPCA(ArrayList<double[]> data, int c) {
		RealMatrix A = MatrixUtils.createRealMatrix(getData(data));
		int m = A.getRowDimension();
		RealMatrix C = extractSubMatrix(A, c);

		// perform pca on C
		EigenDecomposition eigenDecomp = new EigenDecomposition(C.transpose().multiply(C));
		RealMatrix y = eigenDecomp.getV();
		RealMatrix Sigma = eigenDecomp.getD();

		// identify k
		int k = identifyK(Sigma, c);

		// create Sigmak
		RealMatrix Sigmak = Sigma.getSubMatrix(0, k - 1, 0, k - 1);
		RealMatrix Sigmakinverse = Sigmak.copy();

		// create Uk
		RealMatrix Uk = C.createMatrix(m, k);
		for (int i = 0; i < k; i++) {
			Sigmak.setEntry(i, i, Math.sqrt(Sigmak.getRowVector(i).getEntry(i)));
			Sigmakinverse.setEntry(i, i, 1 / Sigmak.getRowVector(i).getEntry(i));
			Uk.setColumnVector(i,C.operate(y.getColumnVector(i)).mapDivide(Sigmak.getRowVector(i).getEntry(i)));
		}

		// create results
		RealMatrix[] ret = new RealMatrix[2];
		ret[0] = Uk.multiply(Sigmak); // AVk
		ret[1] = A.transpose().multiply(Uk.multiply(Sigmakinverse.transpose()));
		return ret;
	}

	private static RealMatrix extractSubMatrix(RealMatrix A, int c) {
		double frobNorm = A.getFrobeniusNorm();
		frobNorm = frobNorm * frobNorm;
		int n = A.getColumnDimension();

		// identify columns with largest variance
		PriorityQueue<Double> pq = new PriorityQueue<Double>();
		double[] colNorms = new double[n];
		double v;
		for (int i = 0; i < n; i++) {
			v = A.getColumnVector(i).getNorm();
			v = v * v;
			v = v / frobNorm;
			colNorms[i] = v;
			if (pq.size() < c) {
				pq.add(v);
			} else {
				if (v > pq.peek()) {
					pq.poll();
					pq.offer(v);
				}
			}
		}

		// identify column ids with largest variance
		double cutoff = pq.peek();
		int[] selectedCols = new int[c];
		int idx = 0;
		for (int i = 0; i < n; i++) {
			if (idx == c) {
				break;
			}

			if (colNorms[i] >= cutoff) {
				selectedCols[idx++] = i;
			}
		}
		selectedCols = refineSelectedCols(selectedCols, colNorms, n);

		// get all row ids
		int m = A.getRowDimension();
		int[] selectedRows = new int[m];
		for (int i = 0; i < m; i++) {
			selectedRows[i] = i;
		}

		// RealMatrix C = A.getSubMatrix(selectedRows, selectedCols);
		RealMatrix C = A.createMatrix(m, c);
		for (int i = 0; i < c; i++) {
			C.setColumnVector(i, A.getColumnVector(selectedCols[i]).mapDivide(Math.sqrt(c * colNorms[selectedCols[i]])));
		}

		return C;
	}
	
	private static int[] refineSelectedCols(int[] selectedCols, double[] colNorms, int n) {
		int c = selectedCols.length;
		Column[] tmp = new Column[c];
		for (int i = 0; i < c; i++) {
			tmp[i] = new Column(selectedCols[i], colNorms[selectedCols[i]]);
		}
		Arrays.sort(tmp, Collections.reverseOrder());
		
		int[] cols = new int[c];
		int idx = 0;
		int k;
		for (int i = 0; i < c; i++) {
			if (idx >= c) {
				break;
			}
			
			k = (int)(n * tmp[i].value);
			for (int j = 0; j < k && idx < c; j++) {
				cols[idx++] = tmp[i].id;
			}
		}
		
		return cols;
	}

	private static int identifyK(RealMatrix Sigma, int c) {
		double percent = Constants.PERCENT;
		double sum = 0;
		for (int i = 0; i < c; i++) {
			sum += Sigma.getRowVector(i).getEntry(i);
		}
		double v = sum;
		int k = c;
		for (int i = c - 1; i >= 0; i--) {
			v = v - Sigma.getRowVector(i).getEntry(i);
			if (v / sum < percent) {
				break;
			}

			k--;
		}

		return k;
	}

	public static void main(String[] args) {
		try {
			String file = args[0];
			int cols = Integer.parseInt(args[1]);
			int c = Integer.parseInt(args[2]);

			ArrayList<Integer> labels = new ArrayList<Integer>();
			ArrayList<double[]> data = CommonUtils.readData(file, cols, labels);
			double[] mean = CommonUtils.meanVector(data);
			CommonUtils.adjustData(data, mean);
			RealMatrix[] ret = fastPCA(data, c);
			int n = ret[0].getColumnDimension();
			RealMatrix[] full = pca(data);
			n = full[0].getColumnDimension();
			System.out.printf("n = %d\n", n);
			System.out.printf("norm = %f\n", full[0].subtract(ret[0]).getFrobeniusNorm());
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
}